@@ -53,6 +53,9 @@ To create a log of net requests
 /******************************************************************************/
 
 const NetFilteringResultCache = class {
+    shelfLife = 15000;
+    extensionOriginURL = vAPI.getURL('/');
+
     constructor() {
         this.pruneTimer = vAPI.defer.create(( ) => {
             this.prune();
@@ -172,9 +175,6 @@ const NetFilteringResultCache = class {
     }
 };
 
-NetFilteringResultCache.prototype.shelfLife = 15000;
-NetFilteringResultCache.prototype.extensionOriginURL = vAPI.getURL('/');
-
 /******************************************************************************/
 
 // Frame stores are used solely to associate a URL with a frame id.
@@ -274,18 +274,16 @@ const FrameStore = class {
     }
 
     static factory(frameURL, parentId = -1) {
-        const entry = FrameStore.junkyard.pop();
-        if ( entry === undefined ) {
-            return new FrameStore(frameURL, parentId);
+        const FS = FrameStore;
+        if ( FS.junkyard.length !== 0 ) {
+            return FS.junkyard.pop().init(frameURL, parentId);
         }
-        return entry.init(frameURL, parentId);
+        return new FS(frameURL, parentId);
     }
+    static junkyard = [];
+    static junkyardMax = 50;
 };
 
-// To mitigate memory churning
-FrameStore.junkyard = [];
-FrameStore.junkyardMax = 50;
-
 /******************************************************************************/
 
 const CountDetails = class {
@@ -312,19 +310,26 @@ const HostnameDetails = class {
     }
     init(hostname) {
         this.hostname = hostname;
+        this.cname = vAPI.net.canonicalNameFromHostname(hostname);
         this.counts.reset();
+        return this;
     }
     dispose() {
-        this.hostname = '';
-        if ( HostnameDetails.junkyard.length < HostnameDetails.junkyardMax ) {
-            HostnameDetails.junkyard.push(this);
-        }
+        const HD = HostnameDetails;
+        if ( HD.junkyard.length >= HD.junkyardMax ) { return; }
+        HD.junkyard.push(this);
     }
+    static factory(hostname) {
+        const HD = HostnameDetails;
+        if ( HD.junkyard.length !== 0 ) {
+            return HD.junkyard.pop().init(hostname);
+        }
+        return new HD(hostname);
+    }
+    static junkyard = [];
+    static junkyardMax = 100;
 };
 
-HostnameDetails.junkyard = [];
-HostnameDetails.junkyardMax = 100;
-
 const HostnameDetailsMap = class extends Map {
     reset() {
         this.clear();
@@ -530,6 +535,21 @@ const PageStore = class {
         return sender.frameURL;
     }
 
+    getFrameAncestorDetails(frameId) {
+        if ( frameId === 0 ) { return []; }
+        const out = [];
+        for (;;) {
+            const frameStore = this.getFrameStore(frameId);
+            if ( frameStore === null ) { break; }
+            const { domain, hostname } = frameStore;
+            if ( hostname !== undefined ) {
+                out.push({ domain, hostname });
+            }
+            frameId = frameStore.parentId;
+        }
+        return out.slice(1);
+    }
+
     // There is no event to tell us a specific subframe has been removed from
     // the main document. The code below will remove subframes which are no
     // longer present in the root document. Removing obsolete subframes is
@@ -544,7 +564,7 @@ const PageStore = class {
             entries = await webext.webNavigation.getAllFrames({
                 tabId: this.tabId
             });
-        } catch(ex) {
+        } catch {
         }
         if ( Array.isArray(entries) === false ) { return; }
         const toKeep = new Set();
@@ -622,7 +642,7 @@ const PageStore = class {
         ) {
             this.hostnameDetailsMap.set(
                 this.tabHostname,
-                new HostnameDetails(this.tabHostname)
+                HostnameDetails.factory(this.tabHostname)
             );
         }
         return this.hostnameDetailsMap;
@@ -700,7 +720,7 @@ const PageStore = class {
             const hostname = journal[i+0];
             let hnDetails = this.hostnameDetailsMap.get(hostname);
             if ( hnDetails === undefined ) {
-                hnDetails = new HostnameDetails(hostname);
+                hnDetails = HostnameDetails.factory(hostname);
                 this.hostnameDetailsMap.set(hostname, hnDetails);
                 this.contentLastModified = now;
             }
@@ -920,8 +940,11 @@ const PageStore = class {
     }
 
     redirectBlockedRequest(fctxt) {
-        const directives = staticNetFilteringEngine.redirectRequest(redirectEngine, fctxt);
-        if ( directives === undefined ) { return; }
+        const directives = staticNetFilteringEngine.redirectRequest(redirectEngine, fctxt) || [];
+        if ( this.urlSkippableResources.has(fctxt.itype) ) {
+            staticNetFilteringEngine.urlSkip(fctxt, true, directives);
+        }
+        if ( directives.length === 0 ) { return; }
         if ( logger.enabled !== true ) { return; }
         fctxt.pushFilters(directives.map(a => a.logData()));
         if ( fctxt.redirectURL === undefined ) { return; }
@@ -932,19 +955,17 @@ const PageStore = class {
     }
 
     redirectNonBlockedRequest(fctxt) {
-        const transformDirectives = staticNetFilteringEngine.transformRequest(fctxt);
-        const pruneDirectives = fctxt.redirectURL === undefined &&
-            staticNetFilteringEngine.hasQuery(fctxt) &&
-            staticNetFilteringEngine.filterQuery(fctxt) ||
-            undefined;
-        if ( transformDirectives === undefined && pruneDirectives === undefined ) { return; }
+        const directives = [];
+        staticNetFilteringEngine.transformRequest(fctxt, directives);
+        if ( staticNetFilteringEngine.hasQuery(fctxt) ) {
+            staticNetFilteringEngine.filterQuery(fctxt, directives);
+        }
+        if ( this.urlSkippableResources.has(fctxt.itype) ) {
+            staticNetFilteringEngine.urlSkip(fctxt, false, directives);
+        }
+        if ( directives.length === 0 ) { return; }
         if ( logger.enabled !== true ) { return; }
-        if ( transformDirectives !== undefined ) {
-            fctxt.pushFilters(transformDirectives.map(a => a.logData()));
-        }
-        if ( pruneDirectives !== undefined ) {
-            fctxt.pushFilters(pruneDirectives.map(a => a.logData()));
-        }
+        fctxt.pushFilters(directives.map(a => a.logData()));
         if ( fctxt.redirectURL === undefined ) { return; }
         fctxt.pushFilter({
             source: 'redirect',
@@ -952,6 +973,19 @@ const PageStore = class {
         });
     }
 
+    skipMainDocument(fctxt, blocked) {
+        const directives = staticNetFilteringEngine.urlSkip(fctxt, blocked);
+        if ( directives === undefined ) { return; }
+        if ( logger.enabled !== true ) { return; }
+        fctxt.pushFilters(directives.map(a => a.logData()));
+        if ( fctxt.redirectURL !== undefined ) {
+            fctxt.pushFilter({
+                source: 'redirect',
+                raw: fctxt.redirectURL
+            });
+        }
+    }
+
     filterCSPReport(fctxt) {
         if (
             sessionSwitches.evaluateZ(
@@ -1006,10 +1040,29 @@ const PageStore = class {
     }
 
     // The caller is responsible to check whether filtering is enabled or not.
-    filterLargeMediaElement(fctxt, size) {
+    filterLargeMediaElement(fctxt, headers) {
         fctxt.filter = undefined;
-
-        if ( this.allowLargeMediaElementsUntil === 0 ) {
+        if ( this.allowLargeMediaElementsUntil === 0 ) { return 0; }
+        if ( sessionSwitches.evaluateZ('no-large-media', fctxt.getTabHostname() ) !== true ) {
+            this.allowLargeMediaElementsUntil = 0;
+            return 0;
+        }
+        // XHR-based streaming is never blocked but we want to prevent autoplay
+        if ( fctxt.itype === fctxt.XMLHTTPREQUEST ) {
+            const ctype = headers.contentType;
+            if ( ctype.startsWith('audio/') || ctype.startsWith('video/') ) {
+                this.largeMediaTimer.on(500);
+            }
+            return 0;
+        }
+        if ( Date.now() < this.allowLargeMediaElementsUntil ) {
+            if ( fctxt.itype === fctxt.MEDIA ) {
+                const sources = this.allowLargeMediaElementsRegex instanceof RegExp
+                    ? [ this.allowLargeMediaElementsRegex.source ]
+                    : [];
+                sources.push('^' + µb.escapeRegex(fctxt.url));
+                this.allowLargeMediaElementsRegex = new RegExp(sources.join('|'));
+            }
             return 0;
         }
         // Disregard large media elements previously allowed: for example, to
@@ -1020,34 +1073,20 @@ const PageStore = class {
         ) {
             return 0;
         }
-        if ( Date.now() < this.allowLargeMediaElementsUntil ) {
-            const sources = this.allowLargeMediaElementsRegex instanceof RegExp
-                ? [ this.allowLargeMediaElementsRegex.source ]
-                : [];
-            sources.push('^' + µb.escapeRegex(fctxt.url));
-            this.allowLargeMediaElementsRegex = new RegExp(sources.join('|'));
-            return 0;
+        // Regardless of whether a media is blocked, we want to prevent autoplay
+        if ( fctxt.itype === fctxt.MEDIA ) {
+            this.largeMediaTimer.on(500);
         }
-        if (
-            sessionSwitches.evaluateZ(
-                'no-large-media',
-                fctxt.getTabHostname()
-            ) !== true
-        ) {
-            this.allowLargeMediaElementsUntil = 0;
-            return 0;
+        const size = headers.contentLength;
+        if ( isNaN(size) ) {
+            return µb.userSettings.largeMediaSize === 0 ? 1 : 0;
         }
-        if ( (size >>> 10) < µb.userSettings.largeMediaSize ) {
-            return 0;
-        }
-
+        if ( (size >>> 10) < µb.userSettings.largeMediaSize ) { return 0; }
         this.largeMediaCount += 1;
         this.largeMediaTimer.on(500);
-
         if ( logger.enabled ) {
             fctxt.filter = sessionSwitches.toLogData();
         }
-
         return 1;
     }
 
@@ -1116,23 +1155,32 @@ const PageStore = class {
         response.blockedResources =
             this.netFilteringCache.lookupAllBlocked(fctxt.getDocHostname());
     }
+
+    cacheableResults = new Set([
+        µb.FilteringContext.SUB_FRAME
+    ]);
+
+    collapsibleResources = new Set([
+        µb.FilteringContext.IMAGE,
+        µb.FilteringContext.MEDIA,
+        µb.FilteringContext.OBJECT,
+        µb.FilteringContext.SUB_FRAME,
+    ]);
+
+    urlSkippableResources = new Set([
+        µb.FilteringContext.IMAGE,
+        µb.FilteringContext.MAIN_FRAME,
+        µb.FilteringContext.MEDIA,
+        µb.FilteringContext.OBJECT,
+        µb.FilteringContext.OTHER,
+        µb.FilteringContext.SUB_FRAME,
+    ]);
+
+    // To mitigate memory churning
+    static junkyard = [];
+    static junkyardMax = 10;
 };
 
-PageStore.prototype.cacheableResults = new Set([
-    µb.FilteringContext.SUB_FRAME,
-]);
-
-PageStore.prototype.collapsibleResources = new Set([
-    µb.FilteringContext.IMAGE,
-    µb.FilteringContext.MEDIA,
-    µb.FilteringContext.OBJECT,
-    µb.FilteringContext.SUB_FRAME,
-]);
-
-// To mitigate memory churning
-PageStore.junkyard = [];
-PageStore.junkyardMax = 10;
-
 /******************************************************************************/
 
 export { PageStore };
